package mdd.client;

import javafx.beans.property.DoubleProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.Event;
import javafx.scene.Group;
import javafx.scene.input.*;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;
import javafx.scene.text.Font;
import javafx.scene.text.Text;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ParseException;

import java.util.Vector;
import java.util.function.DoubleUnaryOperator;

public class Connectable extends Group implements IGUIEventClient ,ChangeListener<Number>{
    private GUIEventHandler _eventHandler;
    private Circle _border;
    private Circle _circle;
    public boolean connected = false;
    public boolean optimizable = false;
    private Module _parent;
    private ConnectableType _type;

    protected Text _label;
    public String Name = "";
    public String ID = "";
    public int Appendix = 0;
    public Vector<Double> Value = new Vector<>();
    public DoubleProperty x = new SimpleDoubleProperty();
    public DoubleProperty y = new SimpleDoubleProperty();


    public Connectable(Module parent, ConnectableType type, JSONObject json){
        _eventHandler = GUIEventHandler.getEventHandler();
        _eventHandler.addEventListener(this);
        _parent = parent;
        _type = type;

        if (json.containsKey("name")){
            Name = (String) json.get("name");
        }
        if (json.containsKey("id")){
            Appendix = (int) json.get("id");
        }
        if (json.containsKey("value")){
            for (Object o: (JSONArray)json.get("value")) {
                Value.add((Double)o);
            }
        }

        ID = Name + Appendix;
        _label = new Text(Name);

        _label.setFont(new Font("Arial", 15));
        _label.getStyleClass().add("text");

        _parent.translateXProperty().addListener(this);
        _parent.translateYProperty().addListener(this);

        _parent.hlayout.translateXProperty().addListener(this);
        _parent.hlayout.translateYProperty().addListener(this);

        _parent.vlayout.translateXProperty().addListener(this);
        _parent.vlayout.translateYProperty().addListener(this);

        if (_type == ConnectableType.INPUT){
            _parent.inputs.layoutXProperty().addListener(this);
            _parent.inputs.layoutYProperty().addListener(this);
        }
        else{
            _parent.outputs.layoutXProperty().addListener(this);
            _parent.outputs.layoutYProperty().addListener(this);
        }

        layoutXProperty().addListener(this);
        layoutYProperty().addListener(this);

        _border = new Circle(0,0, 7);
        _border.setFill(Color.BLACK);
        _circle = new Circle(0,0,5);
        _circle.setFill(Color.WHITE);

        _circle.setOnMouseClicked(e -> {
            if (e.getButton() == MouseButton.SECONDARY) {
                if (!connected) {
                    optimizable = !optimizable;
                    updateColor();
                }
            }

        });

        getChildren().addAll(_border, _circle);

        _circle.addEventHandler(MouseEvent.DRAG_DETECTED, e -> {
            Dragboard db = startDragAndDrop(TransferMode.MOVE);
            ClipboardContent content = new ClipboardContent();
            JSONObject jmsg = generateJSON();
            content.putString(jmsg.toString());
            System.out.println("[Connectable]: started drag with id = " + jmsg.toString());
            db.setContent(content);
            _eventHandler.publishEvent("DRAG_LINE", jmsg);
            e.consume();
        });

        _circle.addEventHandler(MouseEvent.MOUSE_DRAGGED, Event::consume);

        _circle.addEventHandler(DragEvent.DRAG_OVER, e -> {
            Dragboard db = e.getDragboard();
            if (db.hasString()) {
                e.acceptTransferModes(TransferMode.MOVE);
            }
            e.consume();
        });

        _circle.addEventHandler(DragEvent.DRAG_DROPPED, e -> {
            Dragboard db = e.getDragboard();
            boolean success = false;
            if (db.hasString()) {
                JSONParser parser = new JSONParser();
                JSONObject jmsg = null;
                try {
                    jmsg = (JSONObject)parser.parse(db.getString());
                } catch (ParseException parseException) {
                    parseException.printStackTrace();
                }
                JSONArray jsonList = new JSONArray();
                jsonList.add(jmsg);
                jsonList.add(generateJSON());

                ClipboardContent content = new ClipboardContent();
                content.putString(jsonList.toString());
                db.setContent(content);

                System.out.println("Dropped: " + db.getString());
                success = true;
            }
            //e.setDropCompleted(success);
            //e.consume();
        });

        _circle.addEventFilter(MouseEvent.MOUSE_PRESSED, e -> {
            if(e.getButton() == MouseButton.PRIMARY) {
                _eventHandler.publishEvent("MOUSE_PRESSED", generateJSON());
            }
            if(e.getButton() == MouseButton.SECONDARY) {

            }
            //e.consume();
        });
    }

    private void updateColor(){
        if (connected){
            _circle.setFill(Color.YELLOW);
        } else if (optimizable) {
            _circle.setFill(Color.GREEN);
        } else {
            _circle.setFill(Color.WHITE);
        }
    }

    private JSONObject generateJSON(){
        JSONObject json = new JSONObject();
        if (_type == ConnectableType.INPUT){
            json.put("type", "INPUT");
        }else if(_type == ConnectableType.OUTPUT){
            json.put("type", "OUTPUT");
        }else{
            json.put("type", "ANYTHING");
        }

        json.put("ID", Name + Appendix);
        json.put("parent", _parent.ID);

        return json;
    }

    public void setConnected(boolean state){
        if (state != connected){
            connected = state;
            updateColor();
        }
    }

    @Override
    public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
        if (_type == ConnectableType.INPUT) {
            x.setValue(_parent.getTranslateX() + _parent.inputs.getTranslateX() + _parent.inputs.getLayoutX() + _parent.hlayout.getTranslateX() +  _parent.hlayout.getLayoutX()  + _parent.vlayout.getTranslateX() +  _parent.vlayout.getLayoutX() + getLayoutX());
            y.setValue(_parent.getTranslateY() + _parent.inputs.getLayoutY() + _parent.hlayout.getTranslateY() +  _parent.hlayout.getLayoutY() + _parent.vlayout.getTranslateY() + _parent.vlayout.getLayoutY() + getLayoutY());
        }
        else{
            x.setValue(_parent.getTranslateX() + _parent.outputs.getTranslateX() + _parent.outputs.getLayoutX() + _parent.hlayout.getTranslateX() +  _parent.hlayout.getLayoutX() + _parent.vlayout.getTranslateX() + _parent.vlayout.getLayoutX() + getLayoutX());
            y.setValue(_parent.getTranslateY() + _parent.outputs.getLayoutY() + _parent.hlayout.getTranslateY() +  _parent.hlayout.getLayoutY() + _parent.vlayout.getTranslateY() + _parent.vlayout.getLayoutY() + getLayoutY());
        }
    }

    @Override
    public void processGUIEvent(String event, JSONObject args) {

    }

    @Override
    public void close() throws Exception {
        _eventHandler.removeEventListener(this);
    }
}
